"""Defines Intel RAPL related classes."""
import os
import re
from pathlib import Path
from typing import Callable, List, Set

import cpuinfo
import numpy as np

from perun import log
from perun.units import Joule

from .backend import Backend, backend
from .device import Device

RAPL_PATH = "/sys/class/powercap/"

DIR_RGX = r"intel-rapl:(\d)$"
SUBDIR_RGX = r"intel-rapl:\d:\d$"


@backend
class IntelRAPLBackend(Backend):
    """Intel RAPL as a source of cpu and memory devices.

    Uses pyRAPL to gather device information and creates metrics for each available device
    """

    name = "Intel RAPL"
    description = "Reads energy usage from CPUs and DRAM using Intel RAPL"

    def __init__(self) -> None:
        """Init IntelRAPLBackend."""
        super().__init__()
        log.debug("Initialized pyrapl")
        self.setup()

    def setup(self):
        """Check Intel RAPL access."""
        cpuInfo = cpuinfo.get_cpu_info()
        self.cpu_name = cpuInfo["brand_raw"]
        self.thread_count = cpuInfo["count"]

        raplPath = Path(RAPL_PATH)
        self._devices = {}

        if not raplPath.exists():
            raise ImportWarning("No powercap interface")

        for child in raplPath.iterdir():
            match = re.match(DIR_RGX, child.name)
            if match:
                if os.access(child / "energy_uj", os.R_OK):
                    socket = match.groups()[0]
                    device_name = open(child / "name", "r").readline().strip()
                    self._devices[f"{socket}_{device_name}"] = {
                        "max_energy_uj": np.uint64(
                            open(child / "max_energy_range_uj", "r").readline().strip()
                        ),
                        "energy_path": str(child / "energy_uj"),
                    }
                    for grandchild in child.iterdir():
                        match = re.match(SUBDIR_RGX, grandchild.name)
                        if match:
                            device_name = (
                                open(grandchild / "name", "r").readline().strip()
                            )
                            self._devices[f"{socket}_{device_name}"] = {
                                "max_energy_uj": np.uint64(
                                    open(grandchild / "max_energy_range_uj", "r")
                                    .readline()
                                    .strip()
                                ),
                                "energy_path": str(grandchild / "energy_uj"),
                            }

    def close(self) -> None:
        """Backend shutdown code (does nothing for intel rapl)."""
        return

    def visibleDevices(self) -> Set[str]:
        """
        Return string ids of visible devices.

        Returns:
            Set[str]: Set with device string ids
        """
        return set(self._devices.keys())

    def getDevices(self, deviceList: Set[str]) -> List[Device]:
        """
        Gather devive objects based on a set of device ids.

        Args:
            deviceList (Set[str]): Set containing devices ids

        Returns:
            List[Device]: Device objects
        """

        def getCallback(filePath) -> Callable[[], np.number]:
            def func() -> np.number:
                return np.uint64(open(filePath, "r").readline().strip())

            return func

        self.devices: List[Device] = []
        for device in deviceList:
            if device in self._devices:

                deviceInfo = self._devices[device]
                self.devices.append(
                    Device(
                        device,
                        f"{self.cpu_name}:{device}",
                        Joule(),
                        "micro",
                        0,
                        deviceInfo["max_energy_uj"],
                        "uint64",
                        getCallback(deviceInfo["energy_path"]),
                    )
                )

        return self.devices


IntelRAPLBackend()



"""perun module."""
# flake8: noqa
__version__ = "0.1.0-beta.9"

import os

from dotenv import load_dotenv

# SETUP ENV
load_dotenv()
os.environ["IBV_FORK_SAFE"] = "1"
os.environ["RDMAV_FORK_SAFE"] = "1"
log_lvl = os.environ["LOG_LVL"] if "LOG_LVL" in os.environ else "INFO"

from perun.logging import init_logging

log = init_logging(log_lvl)

from perun.perun import getDeviceConfiguration, monitor, perunSubprocess, postprocessing
from perun.report import report
from perun.storage import ExperimentStorage, LocalStorage

__all__ = [
    "perunSubprocess",
    "getDeviceConfiguration",
    "LocalStorage",
    "ExperimentStorage",
    "log",
    "monitor",
    "report",
    "postprocessing",
]



"""Command line api definition.

Uses click https://click.palletsprojects.com/en/8.1.x/ to manage complex cmdline configurations.
"""
from datetime import datetime
from pathlib import Path
from typing import List, Optional

import click

from perun.storage import ExperimentStorage, LocalStorage


@click.group()
def cli():
    """Entry point for the perun command line interface."""


@cli.command()
@click.argument("exp_hdf5", type=click.Path(exists=True))
def postprocess(exp_hdf5: str):
    """
    Apply post-processing to EXP_HDF5 experiment file.

    EXP_HDF5 is an hdf5 file generated by perun after monitoring a script, containing data gathered from hardware devices.
    """
    from mpi4py import MPI

    import perun

    expPath = Path(exp_hdf5)
    expStrg = ExperimentStorage(expPath, MPI.COMM_WORLD)
    perun.postprocessing(expStrg)
    expStrg.close()


@cli.command()
@click.argument("exp_hdf5", type=click.Path(exists=True))
@click.option(
    "-f",
    "--format",
    type=click.Choice(["txt", "yaml", "yml", "json"]),
    default="txt",
    help="report print format",
)
def report(exp_hdf5: str, format: str):
    """Print consumption report from EXP_HDF5 on the command line on the desired format.

    EXP_HDF5 is an hdf5 file generated by perun after monitoring a script, containing data gathered from hardware devices.
    """
    from mpi4py import MPI

    import perun

    expPath = Path(exp_hdf5)
    expStrg = ExperimentStorage(expPath, MPI.COMM_WORLD)
    print(perun.report(expStrg, format=format))
    expStrg.close()


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("script", type=click.Path(exists=True))
@click.argument("script_args", nargs=-1)
@click.option(
    "-f", "--frequency", type=float, default=1.0, help="sampling frequency (in Hz)"
)
@click.option(
    "--format",
    type=click.Choice(["txt", "yaml", "yml", "json"]),
    default="txt",
    help="report print format",
)
@click.option(
    "-o",
    "--outdir",
    type=click.Path(exists=False, dir_okay=True, file_okay=False),
    default="./",
    help="experiment data output directory",
)
@click.option(
    "--horeka/--no-horeka",
    type=bool,
    default=False,
    help="experiment data output directory",
)
def monitor(
    script: str,
    script_args: tuple,
    frequency: float,
    outdir: str,
    format: str,
    horeka: bool,
):
    """
    Gather power consumption from hardware devices while SCRIPT [SCRIPT_ARGS] is running.

    SCRIPT is a path to the python script to monitor, run with arguments SCRIPT_ARGS.
    """
    import sys
    from multiprocessing import Event, Process, Queue

    from mpi4py import MPI

    import perun
    from perun import log
    from perun.backend import backends

    comm = MPI.COMM_WORLD
    start_event = Event()
    stop_event = Event()

    # Setup script arguments
    filePath: Path = Path(script)
    outPath: Path = Path(outdir)
    log.debug(f"Script path: {filePath}")
    argIndex = sys.argv.index(str(filePath))
    sys.argv = sys.argv[argIndex:]
    log.debug(f"Script args: { sys.argv }")

    # Get node devices
    log.debug(f"Backends: {backends}")
    lDeviceIds: List[str] = perun.getDeviceConfiguration(comm, backends)

    for backend in backends:
        backend.close()

    log.debug(f"Rank {comm.rank} - lDeviceIds : {lDeviceIds}")

    # If assigned devices, start subprocess
    if len(lDeviceIds) > 0:
        queue: Queue = Queue()
        perunSP = Process(
            target=perun.perunSubprocess,
            args=[queue, start_event, stop_event, lDeviceIds, frequency],
        )
        perunSP.start()
        start_event.wait()

    # Sync everyone
    comm.barrier()
    start = datetime.now()

    # Start script
    try:
        with open(str(filePath), "r") as scriptFile:
            exec(scriptFile.read(), {"__name__": "__main__"})
    except Exception as e:
        log.error("Failed to open file ", filePath)
        log.error(e)
        stop_event.set()
        return

    stop_event.set()
    stop = datetime.now()

    lStrg: Optional[LocalStorage]
    # Obtain perun subprocess results
    if len(lDeviceIds) > 0:
        perunSP.join()
        perunSP.close()
        lStrg = queue.get()
    else:
        lStrg = None

    # Sync
    comm.barrier()

    # Save raw data to hdf5

    if comm.rank == 0:
        if not outPath.exists():
            outPath.mkdir(parents=True)

    scriptName = filePath.name.replace(filePath.suffix, "")
    resultPath = outPath / f"{scriptName}.hdf5"
    log.debug(f"Result path: {resultPath}")
    expStrg = perun.ExperimentStorage(resultPath, comm)
    if lStrg:
        log.debug("Creating new experiment")
        expId = expStrg.addExperimentRun(lStrg.toDict())
        expStrg.saveDeviceData(expId, lStrg)

        if horeka:
            from perun.extras.horeka import get_horeka_measurements

            get_horeka_measurements(
                comm, outPath, expStrg.experimentName, expId, start, stop
            )
    else:
        expId = expStrg.addExperimentRun(None)

    # Post post-process
    comm.barrier()
    perun.postprocessing(expStorage=expStrg)
    comm.barrier()
    if comm.rank == 0:
        print(perun.report(expStrg, expIdx=expId, format=format))

    comm.barrier()
    expStrg.close()


if __name__ == "__main__":
    cli()



"""Report module."""
import json

import yaml

from perun.storage import ExperimentStorage


def report(
    expStrg: ExperimentStorage,
    format: str = "txt",
    expIdx: int = None,
    rawData: bool = False,
) -> str:
    """
    Generate report from ExperimentStorage object.

    Args:
        expStrg (ExperimentStorage): Storage object
        format (str, optional): Report format. Defaults to "txt".
        expIdx (int, optional): Index of experiment run (only relevant with text format). Defaults to None.
        rawData (bool, optional): If raw data should be included. Defaults to False.

    Raises:
        ValueError: _description_

    Returns:
        str: Report string
    """
    if format == "json":
        return json.dumps(expStrg.toDict(rawData), indent=4)
    elif format == "yml" or format == "yaml":
        return yaml.dump(expStrg.toDict(rawData))
    elif format == "txt":
        return _textReport(expStrg, expIdx)
    else:
        raise ValueError("Invalid report format")


def _textReport(expStrg: ExperimentStorage, expIdx: int = None) -> str:
    """
    Create text report from storage object.

    Args:
        expStrg (ExperimentStorage): Storage object
        expIdx (int, optional): Run to get information from. Defaults to -1 (last one)

    Returns:
        str: txt report
    """
    run = expStrg.getExperimentRun(expIdx if expIdx else -1)
    reportStr = (
        "\n------------------------------------------\n"
        "PERUN REPORT\n"
        "\n"
        f"Experiment name: {expStrg.experimentName}\n"
        f"Run number: {expIdx}\n"
        f"Runtime: {run.attrs['totalRunRuntime_s']:.3f}s\n"
        f"Avg. Power draw: {run.attrs['runAvgPower_kW']:.3f}kW\n"
        f"Energy Used: {run.attrs['totalRunEnergy_kJ']:.3f}kJ\n"
        "\n"
    )

    # Print last run info

    for node_id, node in run.items():
        reportStr += f"Node {node_id}\n"
        for device_id, device in node.items():
            if device_id != "t_ns":
                reportStr += f"    {device_id} - Energy: {device.attrs['totalDeviceEnergy_kJ']:.3f}kJ - AvgPower: {device.attrs['avgDevicePower_kW']:.3f}kW\n"
        reportStr += "\n"

    # Print experiment info
    nRuns = len(expStrg.getExperimentRuns())
    totalEnergy_kJ = expStrg.getRootObject().attrs["totalExpEnergy_kJ"]
    reportStr += f"The script has been run a total {nRuns} times, and has used {totalEnergy_kJ:.3f}kJ in total."
    return reportStr
